#
# File:    SchemaDefReShape.py
# Author:  J. Westbrook
# Date:    4-Aug-2018
#
#
# Updates:
#    15-Aug-2018 jdw add next() method for py2 compatibility
#    16-Aug-2018 jdw add remaining layouts for sliced schema
##
"""
Companion class to reshape data objects produced by SchemaDefDataPrep()

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import copy
import itertools
import logging

logger = logging.getLogger(__name__)


class SliceValues(object):
    """Iterator class for parent slice values.
    """

    def __init__(self, schemaDataDictById, schemaDefObj, sliceFilter):
        self.__sD = schemaDefObj
        spiL = self.__sD.getSliceParentItems(sliceFilter)
        spfL = self.__sD.getSliceParentFilters(sliceFilter)
        vD = {}
        for sp in spiL:
            catId = sp["CATEGORY"]
            atId = sp["ATTRIBUTE"]
            vals = []
            if catId in schemaDataDictById:
                for rowD in schemaDataDictById[catId]:
                    if self.__testFilter(rowD, catId, spfL):
                        vals.append(((catId, atId), rowD[atId]))
            vD[(catId, atId)] = vals
        #
        self.index = 0
        logger.debug("Parent value dict %r", vD.items())
        # Make a list of lists and then get the product
        #
        valueL = []
        for _, vL in vD.items():
            valueL.append(vL)
        self.data = list(itertools.product(*valueL))
        logger.debug("Parent value product list %r", self.data)

    def isEmpty(self):
        return len(self.data) < 1

    def __testFilter(self, rowD, catId, filters):
        ok = True
        for flt in filters:
            if catId != flt["CATEGORY"]:
                ok = False
                break
            if flt["ATTRIBUTE"] in rowD and rowD[flt["ATTRIBUTE"]] in flt["VALUES"]:
                continue
            else:
                ok = False
                break

        return ok

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        try:
            result = self.data[self.index]
            self.index += 1
        except IndexError:
            raise StopIteration
        return result

    def __prev__(self):
        self.index -= 1
        if self.index < 0:
            raise StopIteration
        return self.data[self.index]


class SchemaDefReShape(object):

    """Companion class to reshape data objects produced by SchemaDefDataPrep()
    """

    def __init__(self, schemaDefAccessObj, workPath=None, verbose=True):
        self.__verbose = verbose
        self.__workPath = workPath
        self.__debug = False
        self.__sD = schemaDefAccessObj
        #

    def applyShape(self, schemaDataDictById, styleType="rowwise_by_name", collectionName=None):
        """
        """
        return self.__reshapeSchemaData(schemaDataDictById, styleType=styleType, collectionName=collectionName)

    def applySlicedShape(self, schemaDataDictById, styleType="rowwise_by_name", sliceFilter=None, collectionName=None):
        """
        """
        rL = []
        if sliceFilter:
            rL = []
            sliceIndex = self.__sD.getSliceIndex(sliceFilter)
            logger.debug("Slice index %r", sliceIndex)
            #
            #
            sliceValues = SliceValues(schemaDataDictById, self.__sD, sliceFilter)
            if sliceValues.isEmpty():
                return rL
            #
            flagNew = True
            # JDW - This path is better performing -
            if styleType == "rowwise_by_name_with_cardinality" and flagNew:
                logger.debug("Invoking one-pass slice filter %s", sliceFilter)
                rL = self.__sliceRowwiseByNameWithCardOnePass(schemaDataDictById, sliceFilter, sliceIndex, collectionName=collectionName)
                logger.debug("Completed one-pass slice filter %s", sliceFilter)
            else:
                # JDW - This path works but is not well performing
                for ii, sliceValue in enumerate(sliceValues):
                    logger.debug(" %4d filter %s slice value %r", ii, sliceFilter, sliceValue)
                    rD = self.__reshapeSlicedSchemaData(schemaDataDictById, sliceFilter, sliceValue, sliceIndex, styleType=styleType, collectionName=collectionName)
                    logger.debug("rD keys %s", rD.keys())
                    rL.append(rD)
        else:
            return [self.__reshapeSchemaData(schemaDataDictById, styleType=styleType, collectionName=collectionName)]

        return rL

    def __reshapeSlicedSchemaData(self, schemaDataDictById, sliceFilter, sliceValues, sliceIndex, styleType="rowwise_by_name", collectionName=None):
        """  Reorganize and rename input table data object according to the input style preference:

             Input: schemaDataDictById  (styleType="rowwise_by_id")
                                         dict[<schemaId>]   = [ row1asDict[attributeId]=value,  row2asDict[attribute]=value, .. ]

             Output: rowwise_by_name:     dict[<schemaObjName>] = [ row1Dict[attributeName]=value,  row2dict[], .. ]
                     rowwise_no_name:    dict[<schemaObjName>] = {'attributes': [atName1, atName2,... ], 'data' : [[val1, val2, .. ],[val1, val2,... ]]}
                     columnwise_by_name: dict[<schemaObjName>] = {'atName': [val1, val2,... ], atName2: [val1,val2, ... ], ...}
      rowwise_by_name_with_cardinality:  same as rowwise_byName with special handing of tables with unit cardinality
                                                          dict[<schemaObjName>] = row1Dict[attributeName]=value (singleton row)
        """
        rD = {}
        try:
            if styleType == "rowwise_by_name":
                rD = self.__sliceRowwiseByName(schemaDataDictById, sliceFilter, sliceValues, sliceIndex, collectionName=collectionName)
            elif styleType == "rowwise_by_name_with_cardinality":
                rD = self.__sliceRowwiseByNameWithCard(schemaDataDictById, sliceFilter, sliceValues, sliceIndex, collectionName=collectionName)
            elif styleType == "columnwise_by_name":
                rD = self.__sliceColumnwise(schemaDataDictById, sliceFilter, sliceValues, sliceIndex, collectionName=collectionName)
            elif styleType == "rowwise_no_name":
                rD = self.__shapeRowwiseNoName(schemaDataDictById)
            elif styleType == "rowwise_by_id":
                rD = schemaDataDictById
            else:
                rD = schemaDataDictById
                logger.warning("Unsupported style type %s", styleType)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            rD = schemaDataDictById

        return rD

    def __inSlice(self, schemaId, sliceIndex, rowD, parentVals):
        """ Test if the child values in the input row dictionary equal the corresponding input parents values
        """
        ok = True
        try:
            if schemaId in sliceIndex:
                for parentVal in parentVals:
                    if parentVal[0] in sliceIndex[schemaId]:
                        # cAtId = sliceIndex[schemaId][parentVal[0]]
                        tOk = False
                        for cAtId in sliceIndex[schemaId][parentVal[0]]:
                            if cAtId in rowD and rowD[cAtId] == parentVal[1]:
                                tOk = True
                                break
                        if tOk:
                            continue
                        else:
                            ok = False
                            break
                    else:
                        ok = False
                        break
            else:
                ok = False

        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ok = False
        return ok

    def __sliceRowwiseByNameWithCard(self, schemaDataDictById, sliceFilter, sliceValues, sliceIndex, collectionName=None):
        rD = {}
        attributeExcludeD = self.__sD.getCollectionExcludedAttributes(collectionName, asSchemaIds=True)
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            lExtra = schemaObj.isSliceExtra(sliceFilter)
            if lExtra:
                logger.debug("SchemaId %r slice extra %r", schemaId, lExtra)
            if schemaId not in sliceIndex and not schemaObj.isSliceExtra(sliceFilter):
                continue
            # ------
            schemaObjName = self.__sD.getSchemaName(schemaId)
            #
            iRowDList = schemaDataDictById[schemaId]

            #
            oRowDList = []
            for iRowD in iRowDList:
                if lExtra or self.__inSlice(schemaId, sliceIndex, iRowD, sliceValues):
                    oRowD = {}
                    for atId in iRowD:
                        if (schemaId, atId) in attributeExcludeD:
                            continue
                        oRowD[schemaObj.getAttributeName(atId)] = iRowD[atId]
                    oRowDList.append(oRowD)

            if not oRowDList and not schemaObj.isMandatory():
                logger.debug("Schema id %r row length %r mandatory %r", schemaId, len(oRowDList), schemaObj.isMandatory())
                continue

            if schemaObj.hasSliceUnitCardinality(sliceFilter) and len(oRowDList) == 1:
                rD[schemaObjName] = oRowDList[0]
            else:
                rD[schemaObjName] = oRowDList

        return rD

    def __sliceRowwiseByName(self, schemaDataDictById, sliceFilter, sliceValues, sliceIndex, collectionName=None):
        rD = {}
        attributeExcludeD = self.__sD.getCollectionExcludedAttributes(collectionName, asSchemaIds=True)
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            lExtra = schemaObj.isSliceExtra(sliceFilter)
            if lExtra:
                logger.debug("SchemaId %r slice extra %r", schemaId, lExtra)
            if schemaId not in sliceIndex and not schemaObj.isSliceExtra(sliceFilter):
                continue
            # ------
            schemaObjName = self.__sD.getSchemaName(schemaId)
            #
            iRowDList = schemaDataDictById[schemaId]

            oRowDList = []
            for iRowD in iRowDList:
                if lExtra or self.__inSlice(schemaId, sliceIndex, iRowD, sliceValues):
                    oRowD = {}
                    for atId in iRowD:
                        if (schemaId, atId) in attributeExcludeD:
                            continue
                        oRowD[schemaObj.getAttributeName(atId)] = iRowD[atId]
                    oRowDList.append(oRowD)
            #
            if not oRowDList and not schemaObj.isMandatory():
                logger.debug("Schema id %r row length %r mandatory %r", schemaId, len(oRowDList), schemaObj.isMandatory())
                continue
            rD[schemaObjName] = oRowDList

        return rD

    def __sliceColumnwise(self, schemaDataDictById, sliceFilter, sliceValues, sliceIndex, collectionName=None):
        rD = {}
        attributeExcludeD = self.__sD.getCollectionExcludedAttributes(collectionName, asSchemaIds=True)
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            lExtra = schemaObj.isSliceExtra(sliceFilter)
            if lExtra:
                logger.debug("SchemaId %r slice extra %r", schemaId, lExtra)
            if schemaId not in sliceIndex and not schemaObj.isSliceExtra(sliceFilter):
                continue
            # ------
            schemaObjName = self.__sD.getSchemaName(schemaId)
            iRowDList = schemaDataDictById[schemaId]
            #
            colD = {}
            for iRowD in iRowDList:
                if lExtra or self.__inSlice(schemaId, sliceIndex, iRowD, sliceValues):
                    for atId in iRowD:
                        if (schemaId, atId) in attributeExcludeD:
                            continue
                        atName = schemaObj.getAttributeName(atId)
                        if atName not in colD:
                            colD[atName] = []
                        colD[atName].append(iRowD[atId])

            if not colD and not schemaObj.isMandatory():
                logger.debug("Schema id %r row length %r mandatory %r", schemaId, len(colD), schemaObj.isMandatory())
                continue
            rD[schemaObjName] = colD

        return rD

    def __sliceRowwiseNoName(self, schemaDataDictById, sliceFilter, sliceValues, sliceIndex):
        rD = {}
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            lExtra = schemaObj.isSliceExtra(sliceFilter)
            if lExtra:
                logger.debug("SchemaId %r slice extra %r", schemaId, lExtra)
            if schemaId not in sliceIndex and not schemaObj.isSliceExtra(sliceFilter):
                continue
            # ------
            atIdList = self.__sD.getAttributeIdList(schemaId)
            atNameList = self.__sD.getAttributeNameList(schemaId)
            schemaObjName = self.__sD.getSchemaName(schemaId)
            #
            iRowDList = schemaDataDictById[schemaId]

            #
            oRowList = []
            for iRowD in iRowDList:
                if lExtra or self.__inSlice(schemaId, sliceIndex, iRowD, sliceValues):
                    oRowL = []
                    for atId in atIdList:
                        rVal = iRowD[atId] if atId in iRowD else None
                        oRowL.append(rVal)
                    oRowList.append(oRowL)
            #
            if not oRowList and not schemaObj.isMandatory():
                logger.debug("Schema id %r row length %r mandatory %r", schemaId, len(oRowList), schemaObj.isMandatory())
                continue
            rD[schemaObjName] = {"attributes": atNameList, "data": oRowList}

        return rD

    #
    def __reshapeSchemaData(self, schemaDataDictById, styleType="rowwise_by_name", collectionName=None):
        """  Reorganize and rename input table data object according to the input style preference:

             Input: schemaDataDictById  (styleType="rowwise_by_id")
                                         dict[<schemaId>]   = [ row1asDict[attributeId]=value,  row2asDict[attribute]=value, .. ]

             Output: rowwise_by_name:    dict[<schemaObjName>] = [ row1Dict[attributeName]=value,  row2dict[], .. ]
                     rowwise_no_name:    dict[<schemaObjName>] = {'attributes': [atName1, atName2,... ], 'data' : [[val1, val2, .. ],[val1, val2,... ]]}
                     columnwise_by_name: dict[<schemaObjName>] = {'atName': [val1, val2,... ], atName2: [val1,val2, ... ], ...}
      rowwise_by_name_with_cardinality:  same as rowwise_byName with special handing of tables with unit cardinality
                                                          dict[<schemaObjName>] = row1Dict[attributeName]=value (singleton row)
        """
        rD = {}
        try:
            if styleType == "rowwise_by_name":
                rD = self.__shapeRowwiseByName(schemaDataDictById, collectionName=collectionName)
            elif styleType == "rowwise_by_name_with_cardinality":
                rD = self.__shapeRowwiseByNameWithCard(schemaDataDictById, collectionName=collectionName)
            elif styleType == "columnwise_by_name":
                rD = self.__shapeColumnwise(schemaDataDictById, collectionName=collectionName)
            elif styleType == "rowwise_no_name":
                rD = self.__shapeRowwiseNoName(schemaDataDictById)
            elif styleType == "rowwise_by_id":
                rD = schemaDataDictById
            else:
                rD = schemaDataDictById
                logger.warning("Unsupported style type %s", styleType)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            rD = schemaDataDictById

        return rD

    def __shapeRowwiseByName(self, schemaDataDictById, collectionName=None):
        rD = {}
        attributeExcludeD = self.__sD.getCollectionExcludedAttributes(collectionName, asSchemaIds=True)
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            schemaObjName = self.__sD.getSchemaName(schemaId)
            iRowDList = schemaDataDictById[schemaId]
            oRowDList = []
            for iRowD in iRowDList:
                oRowD = {}
                for atId in iRowD:
                    if (schemaId, atId) in attributeExcludeD:
                        continue
                    oRowD[schemaObj.getAttributeName(atId)] = iRowD[atId]
                oRowDList.append(oRowD)
            rD[schemaObjName] = oRowDList
        return rD

    def __shapeRowwiseByNameWithCard(self, schemaDataDictById, collectionName=None):
        rD = {}
        attributeExcludeD = self.__sD.getCollectionExcludedAttributes(collectionName, asSchemaIds=True)
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            schemaObjName = self.__sD.getSchemaName(schemaId)
            unitCard = self.__sD.hasUnitCardinality(schemaId)
            iRowDList = schemaDataDictById[schemaId]
            #
            if unitCard and len(iRowDList) == 1:
                iRowD = iRowDList[0]
                oRowD = {}
                for atId in iRowD:
                    if (schemaId, atId) in attributeExcludeD:
                        continue
                    oRowD[schemaObj.getAttributeName(atId)] = iRowD[atId]
                rD[schemaObjName] = oRowD
            else:
                oRowDList = []
                for iRowD in iRowDList:
                    oRowD = {}
                    for atId in iRowD:
                        if (schemaId, atId) in attributeExcludeD:
                            continue
                        oRowD[schemaObj.getAttributeName(atId)] = iRowD[atId]
                    oRowDList.append(oRowD)
                rD[schemaObjName] = oRowDList
        return rD

    def __shapeColumnwise(self, schemaDataDictById, collectionName=None):
        rD = {}
        attributeExcludeD = self.__sD.getCollectionExcludedAttributes(collectionName, asSchemaIds=True)
        for schemaId in schemaDataDictById:
            schemaObj = self.__sD.getSchemaObject(schemaId)
            schemaObjName = self.__sD.getSchemaName(schemaId)
            iRowDList = schemaDataDictById[schemaId]
            colD = {}
            for iRowD in iRowDList:
                for atId in iRowD:
                    if (schemaId, atId) in attributeExcludeD:
                        continue
                    atName = schemaObj.getAttributeName(atId)
                    if atName not in colD:
                        colD[atName] = []
                    colD[atName].append(iRowD[atId])
            rD[schemaObjName] = colD
        return rD

    def __shapeRowwiseNoName(self, schemaDataDictById):
        rD = {}
        for schemaId in schemaDataDictById:
            # schemaObj = self.__sD.getSchemaObject(schemaId)
            schemaObjName = self.__sD.getSchemaName(schemaId)
            atIdList = self.__sD.getAttributeIdList(schemaId)
            atNameList = self.__sD.getAttributeNameList(schemaId)
            #
            iRowDList = schemaDataDictById[schemaId]
            oRowList = []
            for iRowD in iRowDList:
                oRowL = []
                for atId in atIdList:
                    rVal = iRowD[atId] if atId in iRowD else None
                    oRowL.append(rVal)
                oRowList.append(oRowL)
            #
            rD[schemaObjName] = {"attributes": atNameList, "data": oRowList}
        return rD

    # ---------------------- ---------------------- ---------------------- ---------------------- ----------------------
    #
    def __sliceRowwiseByNameWithCardOnePass(self, schemaDataDictById, sliceFilter, sliceIndex, collectionName=None):
        debug = False
        #
        attributeExcludeD = self.__sD.getCollectionExcludedAttributes(collectionName, asSchemaIds=True)
        schemaIdExtraL = self.__sD.getSliceExtraSchemaIds(sliceFilter)
        # logger.info("Schema Id extras %r" % schemaIdExtraL)
        #
        for schemaId in schemaDataDictById:
            # sliceIndex = {'schemaId0': {(pCat0,pAt0): chAt0,  (pCat1,pAt1): chAt1}, ... }, ... }
            if schemaId in sliceIndex:
                logger.debug("SLICE %s - schemaId %s -> %r", sliceFilter, schemaId, sliceIndex[schemaId])
        #
        # Build  value -> parent attributes dictionary -
        #
        sliceValues = SliceValues(schemaDataDictById, self.__sD, sliceFilter)
        pvD = {}
        pVals = []
        for pvTupL in sliceValues:
            vL = []
            pL = []
            for pvTup in pvTupL:
                pL.append(pvTup[0])
                vL.append(pvTup[1])
                pVals.append(pvTup[1])
            pvD[tuple(vL)] = tuple(pL)

        retD = {k: {} for k in pVals}
        # Each slice value has the form: (pCat0,pAt0), v0), (pCat1,pAt1), v1),...)
        #
        # Get the dictionary schemaId -> (chAt, chAt) where chAtn are in parent value order -
        #
        sfAtD = {}
        singleKeyAtD = {}
        sliceValues = SliceValues(schemaDataDictById, self.__sD, sliceFilter)
        pvTupL = next(sliceValues)
        #
        # Each parent key in the slice filter -
        ##
        for pvTup in pvTupL:
            logger.debug("VALUES > Slice filter %s parent %r value %r", sliceFilter, pvTup[0], pvTup[1])
            # sliceIndex = {'schemaId0': mD, ... }
            #         mD = {(pCat0,pAt0): chAt0,  (pCat1,pAt1): chAt1}, ... }, ... }
            for schemaId, mD in sliceIndex.items():
                sfAtD[schemaId] = {}
                if schemaId in schemaDataDictById and pvTup[0] in mD:
                    sfAtD[schemaId].setdefault(pvTup[0], []).extend([k for k in mD[pvTup[0]]])
                    singleKeyAtD.setdefault(schemaId, []).extend([k for k in mD[pvTup[0]]])
                if len(sfAtD[schemaId]) > 1:
                    logger.error("Unsupported slice complexity %s key length %d", sliceFilter, len(sfAtD[schemaId]))
        logger.debug("Ordered slice filter attribute dictionary %r", singleKeyAtD)
        #
        # Slice extra objects get replicated in each slice -
        #
        if schemaIdExtraL:
            rD = {}
            for schemaId in schemaIdExtraL:
                schemaObj = self.__sD.getSchemaObject(schemaId)
                schemaObjName = self.__sD.getSchemaName(schemaId)
                iRowDList = schemaDataDictById[schemaId]
                #
                oRowDList = []
                for iRowD in iRowDList:
                    oRowD = {}
                    for atId in iRowD:
                        oRowD[schemaObj.getAttributeName(atId)] = iRowD[atId]
                    oRowDList.append(oRowD)

                if not oRowDList and not schemaObj.isMandatory():
                    logger.debug("Schema id %r row length %r mandatory %r", schemaId, len(oRowDList), schemaObj.isMandatory())
                    continue

                if schemaObj.hasSliceUnitCardinality(sliceFilter) and len(oRowDList) == 1:
                    rD[schemaObjName] = oRowDList[0]
                else:
                    rD[schemaObjName] = oRowDList
            for pv in pVals:
                retD[pv] = copy.copy(rD)
        #
        # logger.info("slice %s retD keys after slice extra insertion %r", sliceFilter, list(retD.keys()))
        # logger.info("slice %s retD after slice extra insertion %r", sliceFilter, retD)

        #
        # Split of the rest by slice value - for single parent key slices  -
        #
        # logger.info("slice %s singleKeyAtD.keys() %r", sliceFilter, list(singleKeyAtD.keys()))
        #
        for schemaId, atL in singleKeyAtD.items():
            #
            # apply collection exclusion filter
            # atL = [atId for atId in atFullL if (schemaId, atId) not in attributeExcludeD]
            #
            schemaObj = self.__sD.getSchemaObject(schemaId)
            # ------
            schemaObjName = self.__sD.getSchemaName(schemaId)
            #
            uFlag = schemaObj.hasSliceUnitCardinality(sliceFilter)
            #
            iRowDList = schemaDataDictById[schemaId]
            rvS = set()
            # logger.info("collectionName %s attributeExcludeD %r", collectionName, attributeExcludeD)
            for iRowD in iRowDList:
                # Add collection exclusion filter here
                oRowD = {schemaObj.getAttributeName(atId): iRowD[atId] for atId in iRowD if (schemaId, atId) not in attributeExcludeD}
                # JDW skip empty rows -
                if not oRowD:
                    continue
                rvS = set([iRowD[at] if at in iRowD else None for at in atL])
                for rv in rvS:
                    # if rv is null or has no parent then the following will fail -
                    # JDW CHANGE - Use a try rather than 'if'   Ignorring error for now
                    try:
                        if uFlag:
                            retD[rv][schemaObjName] = oRowD
                        else:
                            retD[rv].setdefault(schemaObjName, []).append(oRowD)
                    except Exception:
                        pass

            #
            # logger.info("Slice %s Finished loading schema name %s values %r" % (sliceFilter, schemaObjName, rvS))
            #

            #
            if debug:
                for pv in retD:
                    if schemaObjName in retD[pv]:
                        logger.info(">>>>>>>>")
                        logger.info("slice %s schemaObjName %s value %r obj %r", sliceFilter, schemaObjName, pv, retD[pv][schemaObjName])
                        logger.info(">>>>>>>>")
        #
        return list(retD.values())



##
# File:    DataTypeApplicationInfo.py
# Author:  J. Westbrook
# Date:    19-Apr-2018
# Version: 0.001 Initial version
#
# Updates:
#  22-May-2018 jdw standardize data names and fix alignment in default data sections
#  23-May-2018 jdw change assumptions for update method and add tests.
#   7-Jun-2018 jdw rename and rescope.
#  15-Aug-2018 jdw add mapping for JSON types based on generic 'ANY' typing.
#  29-Sep-2018 jdw make JSON date and datetime type explicit in JSON
#  12-Oct-2018 jdw unsuppress datetime mapping
#   7-Jan-2019 jdw applicationName->dataTyping
#
##
"""
Manage mapping of default application data type to dictionary data types.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import copy
import logging

from mmcif.api.DataCategory import DataCategory
from mmcif.api.PdbxContainers import DataContainer

from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class DataTypeApplicationInfo(object):

    cifTypes = [
        "code",
        "ucode",
        "line",
        "uline",
        "text",
        "int",
        "float",
        "name",
        "idname",
        "any",
        "yyyy-mm-dd",
        "uchar3",
        "uchar1",
        "symop",
        "atcode",
        "yyyy-mm-dd:hh:mm",
        "fax",
        "phone",
        "email",
        "code30",
        "float-range",
        "operation_expression",
        "yyyy-mm-dd:hh:mm-flex",
        "ec-type",
        "ucode-alphanum-csv",
        "int-range",
        "point_symmetry",
        "id_list",
        "4x3_matrix",
        "non_negative_int",
        "positive_int",
        "emd_id",
        "pdb_id",
        "point_group",
        "point_group_helical",
        "boolean",
        "author",
        "orcid_id",
        "symmetry_operation",
        "sequence_dep",
        "date_dep",
        "citation_doi",
        "exp_data_doi",
        "asym_id",
        "",
    ]
    # These are generic types -
    appTypes = [
        "char",
        "char",
        "char",
        "char",
        "char",
        "int",
        "float",
        "char",
        "char",
        "text",
        "date",
        "char",
        "char",
        "char",
        "char",
        "datetime",
        "char",
        "char",
        "char",
        "char",
        "char",
        "char",
        "datetime",
        "char",
        "char",
        "char",
        "char",
        "char",
        "char",
        "int",
        "int",
        "char",
        "char",
        "char",
        "char",
        "char",
        "char",
        "char",
        "char",
        "char",
        "char",
        "char",
        "char",
        "char",
        "",
    ]

    defaultWidths = [
        "10",
        "10",
        "80",
        "80",
        "200",
        "10",
        "10",
        "80",
        "80",
        "255",
        "15",
        "4",
        "2",
        "10",
        "6",
        "20",
        "25",
        "25",
        "80",
        "30",
        "30",
        "30",
        "20",
        "10",
        "25",
        "20",
        "80",
        "100",
        "10",
        "10",
        "10",
        "15",
        "20",
        "20",
        "20",
        "5",
        "80",
        "20",
        "80",
        "20",
        "20",
        "20",
        "20",
        "20",
        "",
    ]

    defaultPrecisions = [
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "6",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "0",
        "",
    ]

    #
    def __init__(self, locator=None, dataTyping="ANY", workPath=None, **kwargs):
        self.__workPath = workPath
        self.__locator = locator
        self.__dataTyping = dataTyping
        self.__maxCharWidth = 16382
        self.__dtmD = self.__setup(self.__locator, self.__dataTyping)
        self.__kwargs = kwargs

    def __setup(self, locator, dataTyping):
        appName = "ANY" if dataTyping in ["JSON", "BSON"] else dataTyping
        if locator:
            logger.debug("Using data mapping from %s for %s", locator, appName)
            dtmD = self.readDefaultDataTypeMap(locator, dataTyping=appName)
        else:
            logger.debug(">>>> Falling back to default type mapping. ")
            dtmD = self.getDefaultDataTypeMap(dataTyping=appName)
        # for JSON - transform the generic 'ANY' data types -
        if dataTyping == "JSON":
            for _, tD in dtmD.items():
                if tD["application_name"] == "ANY":
                    if tD["app_type_code"] in ["char", "text"]:
                        tD["app_type_code"] = "string"
                    elif tD["app_type_code"] in ["date", "datetime"]:
                        tD["app_type_code"] = tD["app_type_code"]
                    elif tD["app_type_code"] in ["float"]:
                        tD["app_type_code"] = "number"
                    elif tD["app_type_code"] in ["int"]:
                        tD["app_type_code"] = "integer"
                tD["application_name"] = "JSON"
        elif dataTyping == "BSON":
            for _, tD in dtmD.items():
                if tD["application_name"] == "ANY":
                    if tD["app_type_code"] in ["char", "text"]:
                        tD["app_type_code"] = "string"
                    elif tD["app_type_code"] in ["float", "double"]:
                        tD["app_type_code"] = "double"
                    elif tD["app_type_code"] in ["date", "datetime"]:
                        tD["app_type_code"] = "date"
                    elif tD["app_type_code"] in ["int", "integer"]:
                        tD["app_type_code"] = "int"
                tD["application_name"] = "BSON"
        #
        return dtmD

    def testCache(self):
        return len(self.__dtmD) > 48 and len(DataTypeApplicationInfo.cifTypes) >= 40

    def getDefaultDataTypeMap(self, dataTyping="ANY"):
        try:
            mapD = {}
            for (cifType, simpleType, defWidth, defPrecision) in zip(
                DataTypeApplicationInfo.cifTypes, DataTypeApplicationInfo.appTypes, DataTypeApplicationInfo.defaultWidths, DataTypeApplicationInfo.defaultPrecisions
            ):
                if self.__isNull(cifType):
                    continue
                mapD[cifType] = {
                    "app_type_code": simpleType,
                    "app_precision_default": defPrecision,
                    "app_width_default": defWidth,
                    "type_code": cifType,
                    "application_name": dataTyping,
                }
            return mapD
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return {}

    def updateCharType(self, isKey, appType, dataWidth, defaultWidth, bufferPercent=30.0, minWidth=10):
        """ Skeleton implementation needs to add bracket typing around [80, 256, 1024, 2048, ... ]
        """
        retDataWidth = defaultWidth
        retDataType = appType
        iWidth = dataWidth + int(bufferPercent * 0.01 * dataWidth)
        retDataWidth = iWidth if iWidth > minWidth else minWidth
        if self.__dataTyping.upper() in ["SQL", "MYSQL", "COCKROACH", "CRATE"]:

            if appType.upper() in ["CHAR", "VARCHAR"] and retDataWidth > self.__maxCharWidth:
                retDataType = "TEXT"
            if appType.upper() in ["TEXT"] and isKey:
                retDataType = "CHAR"
        else:
            pass
        return (retDataType, retDataWidth)

    def hasType(self, cifType):
        """
        """
        try:
            return cifType in self.__dtmD
        except Exception:
            return False

    def getAppTypeName(self, cifType, default=None):
        """
        """
        try:
            return self.__dtmD[cifType]["app_type_code"]
        except Exception:
            return default

    def getAppTypeDefaultWidth(self, cifType, default=None):
        """
        """
        try:
            return self.__dtmD[cifType]["app_width_default"]
        except Exception:
            return default

    def getAppTypeDefaultPrecision(self, cifType, default=None):
        """
        """
        try:
            return self.__dtmD[cifType]["app_precision_default"]
        except Exception:
            return default

    def writeDefaultDataTypeMap(self, outPath, dataTyping="ANY"):
        """ Write data file containing application default dictionary to application data type mapping

                  data_rcsb_data_type_map
                    loop_
                    _pdbx_data_type_application_map.application_name
                    _pdbx_data_type_application_map.type_code
                    _pdbx_data_type_application_map.app_type_code
                    _pdbx_data_type_application_map.app_precision_default
                    _pdbx_data_type_application_map.app_width_default
                    # .... type mapping data ...
        """
        try:
            #
            containerList = []
            curContainer = DataContainer("rcsb_data_type_map")
            aCat = DataCategory("pdbx_data_type_application_map")
            aCat.appendAttribute("application_name")
            aCat.appendAttribute("type_code")
            aCat.appendAttribute("app_type_code")
            aCat.appendAttribute("app_width_default")
            aCat.appendAttribute("app_precision_default")
            for (cifType, simpleType, defWidth, defPrecision) in zip(
                DataTypeApplicationInfo.cifTypes, DataTypeApplicationInfo.appTypes, DataTypeApplicationInfo.defaultWidths, DataTypeApplicationInfo.defaultPrecisions
            ):
                if self.__isNull(cifType):
                    continue
                aCat.append([dataTyping, cifType, simpleType, defWidth, defPrecision])
            curContainer.append(aCat)
            containerList.append(curContainer)
            #
            mU = MarshalUtil(workPath=self.__workPath)
            ok = mU.doExport(outPath, containerList, fmt="mmcif", enforceAscii=True, useCharRefs=True, raiseExceptions=True)

            return ok
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def updateDefaultDataTypeMap(self, filePath, mapD, dataTyping="ANY"):
        """ Update data file containing application default data type mapping with any
            updates from the input type mapping dictionary

           mapD['cif_type_code'] -> ['application_name', 'app_type_code', 'app_precision_default', 'app_width_default', 'type_code']

                  data_rcsb_data_type_map
                    loop_
                    _pdbx_data_type_application_map.application_name
                    _pdbx_data_type_application_map.type_code
                    _pdbx_data_type_application_map.app_type_code
                    _pdbx_data_type_application_map.app_precision_default
                    _pdbx_data_type_application_map.app_width_default
                    # .... type mapping data ...
        """
        try:
            #
            mD = copy.deepcopy(mapD)
            mU = MarshalUtil(workPath=self.__workPath)
            containerList = mU.doImport(filePath, fmt="mmcif", enforceAscii=True, useCharRefs=True, raiseExceptions=True)
            for container in containerList:
                if container.getName() == "rcsb_data_type_map":
                    catObj = container.getObj("pdbx_data_type_application_map")
                    rIL = []
                    for ii in range(catObj.getRowCount()):
                        dD = catObj.getRowAttributeDict(ii)
                        if dD["application_name"] == dataTyping:
                            rIL.append(ii)
                            mD[dD["type_code"]] = {k: dD[k] for k in ["application_name", "app_type_code", "app_precision_default", "app_width_default", "type_code"]}
                            continue
                    ok = catObj.removeRows(rIL)
                    atNameL = catObj.getAttributeList()
                    for ky in mapD:
                        row = [mapD[ky][atN] for atN in atNameL]
                        catObj.append(row)
            #
            # Write updated data file
            mU = MarshalUtil(workPath=self.__workPath)
            ok = mU.doExport(filePath, containerList, fmt="mmcif", enforceAscii=True, useCharRefs=True, raiseExceptions=True)

            return ok
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def readDefaultDataTypeMap(self, locator, dataTyping="ANY"):
        """ Read data file containing application default data type mapping

                  data_rcsb_data_type_map
                    loop_
                    _pdbx_data_type_application_map.application_name
                    _pdbx_data_type_application_map.type_code
                    _pdbx_data_type_application_map.app_type_code
                    _pdbx_data_type_application_map.app_precision_default
                    _pdbx_data_type_application_map.app_width_default
                    # .... type mapping data ...

            Return (dict):  map[cifType] -> appType, width, precision
                        mapD['cif_type_code'] -> ['application_name', 'app_type_code', 'app_precision_default', 'app_width_default', 'type_code']
        """
        try:
            #
            mapD = {}
            mU = MarshalUtil(workPath=self.__workPath)
            containerList = mU.doImport(locator, fmt="mmcif", enforceAscii=True, useCharRefs=True, raiseExceptions=True)

            for container in containerList:
                if container.getName() == "rcsb_data_type_map":
                    catObj = container.getObj("pdbx_data_type_application_map")
                    for ii in range(catObj.getRowCount()):
                        dD = catObj.getRowAttributeDict(ii)
                        if dD["application_name"] == dataTyping:
                            mapD[dD["type_code"]] = {k: dD[k] for k in ["app_type_code", "application_name", "type_code"]}
                            mapD[dD["type_code"]].update({k: int(dD[k]) for k in ["app_precision_default", "app_width_default"]})
            return mapD
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return {}

    def __isNull(self, value):
        if not value:
            return True
        if (not value) or (value == "?") or (value == "."):
            return True
        return False



##
# File:    DictMethodResourceProvider.py
# Author:  J. Westbrook
# Date:    3-Jun-2019
# Version: 0.001 Initial version
#
#
# Updates:
#  17-Jul-2019 jdw add resource for common utilities and dictionary api
#   7-Aug-2019 jdw use dictionary locator map
#  13-Aug-2019 jdw return class instances in all cases. Add cache management support.
#   9-Sep-2019 jdw add AtcProvider() and SiftsSummaryProvider()
##
"""
Resource provider for DictMethodHelper tools.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os

from rcsb.db.define.DictionaryApiProviderWrapper import DictionaryApiProviderWrapper
from rcsb.db.helpers.DictMethodCommonUtils import DictMethodCommonUtils
from rcsb.utils.chemref.AtcProvider import AtcProvider
from rcsb.utils.chemref.ChemCompModelProvider import ChemCompModelProvider
from rcsb.utils.chemref.DrugBankProvider import DrugBankProvider
from rcsb.utils.ec.EnzymeDatabaseProvider import EnzymeDatabaseProvider
from rcsb.utils.io.SingletonClass import SingletonClass
from rcsb.utils.seq.SiftsSummaryProvider import SiftsSummaryProvider
from rcsb.utils.struct.CathClassificationProvider import CathClassificationProvider
from rcsb.utils.struct.ScopClassificationProvider import ScopClassificationProvider
from rcsb.utils.taxonomy.TaxonomyProvider import TaxonomyProvider
from rcsb.utils.validation.ValidationReportProvider import ValidationReportProvider

logger = logging.getLogger(__name__)


class DictMethodResourceProvider(SingletonClass):
    """ Resource provider for DictMethodHelper tools.

    """

    def __init__(self, cfgOb, **kwargs):
        """ Resource provider for dictionary method runner.

        Arguments:
            cfgOb {object} -- instance ConfigUtils class

        Keyword agruments:
            configName {string} -- configuration section name (default: default section name)
            cachePath {str} -- path used for temporary file management (default: '.')

        """
        self.__cfgOb = cfgOb

        self.__configName = kwargs.get("configName", self.__cfgOb.getDefaultSectionName())
        self.__cachePath = kwargs.get("cachePath", ".")
        #
        self.__drugBankMappingDict = {}
        self.__csdModelMappingDict = {}
        self.__taxU = None
        self.__ecU = None
        self.__scopU = None
        self.__cathU = None
        self.__dbU = None
        self.__ccmU = None
        self.__commonU = None
        self.__dApiW = None
        self.__atcP = None
        self.__siftsAbbreviated = kwargs.get("siftsAbbreviated", "PROD")
        self.__ssP = None
        self.__vrptP = None
        #
        #
        # self.__wsPattern = re.compile(r"\s+", flags=re.UNICODE | re.MULTILINE)
        # self.__re_non_digit = re.compile(r"[^\d]+")
        #
        self.__resourcesD = {
            "SiftsSummaryProvider instance": self.__fetchSiftsSummaryProvider,
            "Dictionary API instance (pdbx_core)": self.__fetchDictionaryApi,
            "TaxonomyProvider instance": self.__fetchTaxonomyProvider,
            "ScopProvider instance": self.__fetchScopProvider,
            "CathProvider instance": self.__fetchCathProvider,
            "EnzymeProvider instance": self.__fetchEnzymeProvider,
            "DrugBankProvider instance": self.__fetchDrugBankProvider,
            "ChemCompModelProvider instance": self.__fetchChemCompModelProvider,
            "AtcProvider instance": self.__fetchAtcProvider,
            "DictMethodCommonUtils instance": self.__fetchCommonUtils,
            "ValidationProvider instance": self.__fetchValidationProvider,
        }
        logger.debug("Dictionary resource provider init completed")
        #

    def echo(self, msg):
        logger.info(msg)

    def getResource(self, resourceName, default=None, useCache=True, **kwargs):
        """ Return the named input resource or the default value.

        Arguments:
            resourceName {str} -- resource name
            useCache (bool, optional): use current cace. Defaults to True.

        Keyword Arguments:
            default {obj} -- default return value for missing resources (default: {None})

        Returns:
            [obj] -- resource object
        """
        logger.debug("Requesting resource %r", resourceName)
        if resourceName in self.__resourcesD:
            return self.__resourcesD[resourceName](self.__cfgOb, self.__configName, self.__cachePath, useCache=useCache, **kwargs)
        else:
            logger.error("Request for unsupported resource %r returning %r", resourceName, default)
        #
        return default

    def cacheResources(self, useCache=False, **kwargs):
        """Update and optionally clear all resource caches.

        Args:
            useCache (bool, optional): use current cace. Defaults to False.

        Returns:
            bool: True for success or False otherwise
        """
        ret = True
        for resourceName in self.__resourcesD:
            logger.info("Caching resources for %r", resourceName)
            tU = self.__resourcesD[resourceName](self.__cfgOb, self.__configName, self.__cachePath, useCache=useCache, **kwargs)
            ok = tU.testCache()
            ret = ret and ok
            logger.info("After %r status %r of %r", resourceName, ok, ret)
        return ret

    def __fetchTaxonomyProvider(self, cfgOb, configName, cachePath, useCache=True, **kwargs):
        logger.debug("configName %s cachePath %s kwargs %r", configName, cachePath, kwargs)
        if not self.__taxU:
            taxonomyDataPath = os.path.join(cachePath, cfgOb.get("NCBI_TAXONOMY_CACHE_DIR", sectionName=configName))
            self.__taxU = TaxonomyProvider(taxDirPath=taxonomyDataPath, useCache=useCache, **kwargs)
        return self.__taxU

    def __fetchScopProvider(self, cfgOb, configName, cachePath, useCache=True, **kwargs):
        logger.debug("configName %s cachePath %s kwargs %r", configName, cachePath, kwargs)
        if not self.__scopU:
            structDomainDataPath = os.path.join(cachePath, cfgOb.get("STRUCT_DOMAIN_CLASSIFICATION_CACHE_DIR", sectionName=configName))
            self.__scopU = ScopClassificationProvider(scopDirPath=structDomainDataPath, useCache=useCache, **kwargs)
        return self.__scopU

    def __fetchCathProvider(self, cfgOb, configName, cachePath, useCache=True, **kwargs):
        logger.debug("configName %s cachePath %s kwargs %r", configName, cachePath, kwargs)
        if not self.__cathU:
            structDomainDataPath = os.path.join(cachePath, cfgOb.get("STRUCT_DOMAIN_CLASSIFICATION_CACHE_DIR", sectionName=configName))
            self.__cathU = CathClassificationProvider(cathDirPath=structDomainDataPath, useCache=useCache, **kwargs)
        return self.__cathU

    def __fetchEnzymeProvider(self, cfgOb, configName, cachePath, useCache=True, **kwargs):
        logger.debug("configName %s cachePath %s kwargs %r", configName, cachePath, kwargs)
        if not self.__ecU:
            enzymeDataPath = os.path.join(cachePath, cfgOb.get("ENZYME_CLASSIFICATION_CACHE_DIR", sectionName=configName))
            self.__ecU = EnzymeDatabaseProvider(enzymeDirPath=enzymeDataPath, useCache=useCache, **kwargs)
        return self.__ecU

    #
    def __fetchDrugBankProvider(self, cfgOb, configName, cachePath, useCache=True, **kwargs):
        logger.debug("configName %s cachePath %s kwargs %r", configName, cachePath, kwargs)
        if not self.__dbU:
            dbDataPath = os.path.join(cachePath, cfgOb.get("DRUGBANK_CACHE_DIR", sectionName=configName))
            un = cfgOb.get("_DRUGBANK_AUTH_USERNAME", sectionName=configName)
            pw = cfgOb.get("_DRUGBANK_AUTH_PASSWORD", sectionName=configName)
            self.__dbU = DrugBankProvider(dirPath=dbDataPath, useCache=useCache, username=un, password=pw, **kwargs)
        return self.__dbU

    def __fetchChemCompModelProvider(self, cfgOb, configName, cachePath, useCache=True, **kwargs):
        logger.debug("configName %s cachePath %s kwargs %r", configName, cachePath, kwargs)
        if not self.__ccmU:
            dirPath = os.path.join(cachePath, cfgOb.get("CHEM_COMP_MODEL_CACHE_DIR", sectionName=configName))
            self.__ccmU = ChemCompModelProvider(dirPath=dirPath, useCache=useCache, **kwargs)
        return self.__ccmU

    def __fetchAtcProvider(self, cfgOb, configName, cachePath, useCache=True, **kwargs):
        logger.debug("configName %s cachePath %s kwargs %r", configName, cachePath, kwargs)
        if not self.__atcP:
            dirPath = os.path.join(cachePath, cfgOb.get("ATC_CACHE_DIR", sectionName=configName))
            self.__atcP = AtcProvider(dirPath=dirPath, useCache=useCache, **kwargs)
        return self.__atcP

    def __fetchSiftsSummaryProvider(self, cfgOb, configName, cachePath, useCache=True, **kwargs):
        logger.debug("configName %s cachePath %s kwargs %r", configName, cachePath, kwargs)
        if not self.__ssP:
            srcDirPath = os.path.join(cachePath, cfgOb.getPath("SIFTS_SUMMARY_DATA_PATH", sectionName=configName))
            cacheDirPath = os.path.join(cachePath, cfgOb.get("SIFTS_SUMMARY_CACHE_DIR", sectionName=configName))
            logger.debug("ssP %r %r", srcDirPath, cacheDirPath)
            self.__ssP = SiftsSummaryProvider(srcDirPath=srcDirPath, cacheDirPath=cacheDirPath, useCache=useCache, abbreviated=self.__siftsAbbreviated, **kwargs)
            logger.debug("ssP entry count %d", self.__ssP.getEntryCount())
        return self.__ssP

    def __fetchValidationProvider(self, cfgOb, configName, cachePath, useCache=True, **kwargs):
        logger.debug("configName %s cachePath %s kwargs %r", configName, cachePath, kwargs)
        if not self.__vrptP:
            dirPath = os.path.join(cachePath, cfgOb.get("DICTIONARY_CACHE_DIR", sectionName=configName))
            self.__vrptP = ValidationReportProvider(dirPath=dirPath, useCache=useCache)
        #
        return self.__vrptP

    def __fetchCommonUtils(self, cfgOb, configName, cachePath, useCache=None, **kwargs):
        logger.debug("configName %s cachePath %r kwargs %r", configName, cachePath, kwargs)
        _ = cfgOb
        _ = useCache
        if not self.__commonU:
            self.__commonU = DictMethodCommonUtils(**kwargs)
        return self.__commonU

    def __fetchDictionaryApi(self, cfgOb, configName, cachePath, useCache=None, **kwargs):
        logger.debug("configName %s cachePath %s kwargs %r", configName, cachePath, kwargs)
        schemaName = kwargs.get("schemaName", "pdbx_core")
        self.__dApiW = DictionaryApiProviderWrapper(cfgOb, cachePath, useCache=useCache)
        dictApi = self.__dApiW.getApiByName(schemaName)
        # numRev = dictApi.getDictionaryRevisionCount()
        return dictApi



##
# File:    DictionaryApiProviderWrapper.py
# Author:  J. Westbrook
# Date:   18-Aug-2019
# Version: 0.001 Initial version
#
# Updates:
#
##
"""
Wrapper for dictionary API provider.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os.path

from rcsb.db.define.DictionaryApiProvider import DictionaryApiProvider
from rcsb.utils.io.SingletonClass import SingletonClass

logger = logging.getLogger(__name__)


class DictionaryApiProviderWrapper(SingletonClass):
    """ Wrapper for dictionary API provider.
    """

    def __init__(self, cfgOb, cachePath, useCache=True, **kwargs):
        """Wrapper for dictionary API provider.

        Args:
            cfgOb (object):  ConfigInfo() object instance
            cachePath (str): top path to contain the dictionary cache directory
            useCache (bool, optional): flag to use cached files. Defaults to True.

        """
        self.__cfgOb = cfgOb
        self.__configName = self.__cfgOb.getDefaultSectionName()
        self.__contentInfoConfigName = "content_info_helper_configuration"
        self.__dictLocatorMap = self.__cfgOb.get("DICT_LOCATOR_CONFIG_MAP", sectionName=self.__contentInfoConfigName)
        dirPath = os.path.join(cachePath, self.__cfgOb.get("DICTIONARY_CACHE_DIR", sectionName=self.__configName))
        self.__dP = DictionaryApiProvider(dirPath, useCache=useCache, **kwargs)
        logger.debug("Leaving constructor")

    def getApiByLocators(self, dictLocators, **kwargs):
        """Return a dictionary API object for the input dictionary locator list.

        Args:
            dictLocators (list str): list of dictionary locators

        Returns:
            (object): Instance of DictionaryApi()
        """
        return self.__dP.getApi(dictLocators, **kwargs)

    def getApiByName(self, databaseName, **kwargs):
        """Return a dictionary API object for the input schema name.

        Args:
            databaseName (str): database schema name

        Returns:
            (object): Instance of DictionaryApi()
        """
        if databaseName not in self.__dictLocatorMap:
            logger.error("Missing dictionary locator configuration for database schema %s", databaseName)
            dictLocators = []
        else:
            dictLocators = [self.__cfgOb.getPath(configLocator, sectionName=self.__contentInfoConfigName) for configLocator in self.__dictLocatorMap[databaseName]]
        #
        return self.__dP.getApi(dictLocators, **kwargs)



##
# File:    PdbxSchamaMapReader.py
# Author:  J. Westbrook
# Date:    4-Jan-2013
#
# Updates:
#  7-Jan-2013 jdw Revised organization of instance data item mapping to attributes.
#  9-Jan-2013 jdw revise handling of mege index.
# 11=Jan-2013 jdw add table and attribute abbreviation support.
# 13-Jan-2013 jdw add test for mysql  index size maximum
# 23-Mar-2018 jdw cleanup logging
##
"""
Example reader for RCSB schema map data files exporting the data structure used by the
rcsb.db.schema.SchemaMapDef class hierarchy.

This modules provides some backward access to prior schema versions used by legacy loader tools.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging

from mmcif.api.PdbxContainers import CifName
from mmcif.io.PdbxReader import PdbxReader

logger = logging.getLogger(__name__)


class PdbxSchemaMapReader(object):
    def __init__(self, verbose=True):
        self.__verbose = verbose
        self.__pathSchemaMapFile = None
        self.__tableNameList = []
        self.__atDefList = []
        self.__atMapList = []
        self.__tableAbbrev = {}
        self.__attribAbbrev = {}

    def read(self, schemaMapFile):
        self.__pathSchemaMapFile = schemaMapFile
        self.__tableNameList, self.__atDefList, self.__atMapList, self.__tableAbbrev, self.__attribAbbrev = self.__readSchemaMap(schemaMapFile)
        return True

    def dump(self, ofh):
        ofh.write("Table name list: %s\n" % self.__tableNameList)
        for dD in self.__atDefList:
            ofh.write("\n\nAttribute def: %s\n" % dD.items())

        for dD in self.__atMapList:
            ofh.write("\n\nAttribute map: %s\n" % dD.items())

        for k, val in self.__tableAbbrev.items():
            ofh.write("Table %s - abbreviation %s\n" % (k, val))

        for tN, dD in self.__attribAbbrev.items():
            for k, val in dD.items():
                ofh.write("Table %s - attribute %s  abbreviation %s\n" % (tN, k, val))

    def __convertDataType(self, aType, aWidth=0, aPrecision=0):
        _ = aPrecision
        if aType.lower() == "char" or aType.lower() == "varchar":
            if aWidth < 65000:
                retType = "VARCHAR"
            else:
                retType = "TEXT"
        elif aType.lower() == "int":
            retType = "INT"
        elif aType.lower() == "float":
            retType = "FLOAT"
        elif aType.lower() == "date" or aType.lower() == "datetime":
            retType = aType.upper()
        else:
            retType = None
            logger.error("+ERROR - UNKNOWN DATA TYPE %s", aType)

        return retType

    def __toBool(self, flag):
        if flag.lower() == "y" or flag.lower() == "yes" or flag == "1":
            return True
        else:
            return False

    def __getTableAbbrev(self, tableName):
        if tableName in self.__tableAbbrev:
            return self.__tableAbbrev[tableName]
        else:
            return tableName

    def __getAttributeAbbrev(self, tableName, attributeName):
        try:
            return self.__attribAbbrev[tableName][attributeName]
        except Exception:
            return attributeName

    def makeSchemaDef(self):
        sD = {}
        for tableName in self.__tableNameList:
            if tableName in ["rcsb_columninfo", "columninfo", "tableinfo", "rcsb_tableinfo"]:
                continue
            dD = {}
            tableAbbrev = self.__getTableAbbrev(tableName)
            tU = tableAbbrev.upper()
            dD["SCHEMA_ID"] = tU
            dD["SCHEMA_NAME"] = tableAbbrev
            dD["SCHEMA_TYPE"] = "transactional"
            dD["ATTRIBUTES"] = {}
            dD["ATTRIBUTE_INFO"] = {}
            dD["ATTRIBUTE_MAP"] = {}
            #
            # create a sub list for this table -
            infoL = []
            for atD in self.__atDefList:
                if atD["table_name"] == tableName:
                    infoL.append(atD)
            #
            mapD = {}
            for atD in self.__atMapList:
                if atD["target_table_name"] == tableName:
                    attributeName = atD["target_attribute_name"]
                    attributeAbbrev = self.__getAttributeAbbrev(tableName, attributeName)
                    atU = attributeAbbrev.upper()
                    itN = atD["source_item_name"] if atD["source_item_name"] not in ["?", "."] else None
                    if itN is not None:
                        catNameM = CifName.categoryPart(itN)
                        attNameM = CifName.attributePart(itN)
                    else:
                        catNameM = None
                        attNameM = None

                    # cId = atD['condition_id'] if atD['condition_id'] not in ['?', '.'] else None
                    fId = atD["function_id"] if atD["function_id"] not in ["?", "."] else None
                    if fId is not None and catNameM is None:
                        mapD[atU] = (catNameM, attNameM, fId, None)
                    else:
                        mapD[atU] = (catNameM, attNameM, fId, None)

            #
            try:
                indexList = []
                for (ii, atD) in enumerate(infoL):
                    attributeName = atD["attribute_name"]
                    attributeAbbrev = self.__getAttributeAbbrev(tableName, attributeName)
                    atU = attributeAbbrev.upper()
                    #
                    td = {}
                    # 'data_type','index_flag','null_flag','width','precision','populated'
                    td["APP_TYPE"] = self.__convertDataType(atD["data_type"], aWidth=int(atD["width"]))
                    td["WIDTH"] = int(atD["width"])
                    td["PRECISION"] = int(atD["precision"])
                    td["NULLABLE"] = not self.__toBool(atD["null_flag"])
                    td["PRIMARY_KEY"] = self.__toBool(atD["index_flag"])
                    td["ORDER"] = ii + 1
                    if td["PRIMARY_KEY"]:
                        indexList.append(atU)
                    dD["ATTRIBUTES"][atU] = attributeAbbrev
                    dD["ATTRIBUTE_INFO"][atU] = td
                    dD["ATTRIBUTE_MAP"][atU] = mapD[atU]
            except Exception as e:
                logger.error("Failing for table %r attribute %r", tableName, attributeName)
                logger.exception("Failing with %s", str(e))

            #
            if self.__verbose and len(indexList) > 16:
                logger.debug("+WARNING - %s index list exceeds MySQL max length %d", tableName, len(indexList))
            mergeDict = {}
            deleteAttributeList = []
            for atU in indexList:
                tN = dD["ATTRIBUTE_MAP"][atU][0]
                aN = dD["ATTRIBUTE_MAP"][atU][1]
                fN = dD["ATTRIBUTE_MAP"][atU][2]
                if aN is not None:
                    if tN not in mergeDict:
                        mergeDict[tN] = []
                    mergeDict[tN].append(aN)
                #
                # Using RCSB convention of including one attribute in each table corresponding to the datablockId()
                #   this attributeId is used a key pre-insert deletions.
                #
                if fN in ["datablockid()"]:
                    deleteAttributeList.append(atU)
            #
            # Assign a merge index to this instance category
            #
            for k, v in mergeDict.items():
                dD["MAP_MERGE_INDICES"] = {k: {"TYPE": "EQUI-JOIN", "ATTRIBUTES": tuple(v)}}

            if deleteAttributeList:
                dD["SCHEMA_DELETE_ATTRIBUTE"] = deleteAttributeList[0]
                dD["INDICES"] = {"p1": {"TYPE": "UNIQUE", "ATTRIBUTES": tuple(indexList)}, "s1": {"TYPE": "SEARCH", "ATTRIBUTES": tuple(deleteAttributeList)}}
            else:
                dD["INDICES"] = {"p1": {"TYPE": "UNIQUE", "ATTRIBUTES": tuple(indexList)}}
                logger.debug("+WARNING - No delete attribute for table %s", tableName)

            if not mergeDict:
                logger.debug("+WARNING - No merge index possible for table %s", tableName)

            sD[tU] = dD
        return sD

    def __readSchemaMap(self, schemaMapFile):
        """Read RCSB schema map file and return the list of table names, attribute definitions,
           attribute mapping, table and attribute abbreviations.
        """
        tableNameList = []
        atDefList = []
        atMapList = []
        tableAbbrevD = {}
        attribAbbrevD = {}
        try:
            #
            myContainerList = []
            ifh = open(schemaMapFile, "r")
            pRd = PdbxReader(ifh)
            pRd.read(myContainerList)
            ifh.close()
            #
            for myContainer in myContainerList:
                cN = str(myContainer.getName()).lower()
                #
                # read schema details --
                #
                if cN == "rcsb_schema":
                    #
                    catObj = myContainer.getObj("rcsb_table")
                    if catObj is not None:
                        i1 = catObj.getAttributeIndex("table_name")
                        for row in catObj.getRowList():
                            tableNameList.append(row[i1])
                    #
                    catObj = myContainer.getObj("rcsb_attribute_def")
                    atList = ["table_name", "attribute_name", "data_type", "index_flag", "null_flag", "width", "precision", "populated"]
                    indList = []
                    if catObj is not None:
                        for at in atList:
                            indList.append(catObj.getAttributeIndex(at))
                        for row in catObj.getRowList():
                            dD = {}
                            for ii, at in enumerate(atList):
                                dD[at] = row[indList[ii]]
                            atDefList.append(dD)
                    #
                    # _rcsb_table_abbrev.table_name
                    # _rcsb_table_abbrev.table_abbrev
                    #
                    catObj = myContainer.getObj("rcsb_table_abbrev")
                    if catObj is not None:
                        i1 = catObj.getAttributeIndex("table_name")
                        i2 = catObj.getAttributeIndex("table_abbrev")
                        for row in catObj.getRowList():
                            tableAbbrevD[row[i1]] = row[i2]
                    #
                    # _rcsb_attribute_abbrev.table_name
                    # _rcsb_attribute_abbrev.attribute_name
                    # _rcsb_attribute_abbrev.attribute_abbrev

                    catObj = myContainer.getObj("rcsb_attribute_abbrev")
                    if catObj is not None:
                        i1 = catObj.getAttributeIndex("table_name")
                        i2 = catObj.getAttributeIndex("attribute_name")
                        i3 = catObj.getAttributeIndex("attribute_abbrev")
                        for row in catObj.getRowList():
                            if row[i1] not in attribAbbrevD:
                                attribAbbrevD[row[i1]] = {}
                            attribAbbrevD[row[i1]][row[i2]] = row[i3]

                # read attribute mapping details --
                #
                elif cN == "rcsb_schema_map":
                    catObj = myContainer.getObj("rcsb_attribute_map")
                    atList = ["target_table_name", "target_attribute_name", "source_item_name", "condition_id", "function_id"]
                    indList = []
                    if catObj is not None:
                        for at in atList:
                            indList.append(catObj.getAttributeIndex(at))
                        for row in catObj.getRowList():
                            dD = {}
                            for ii, at in enumerate(atList):
                                dD[at] = row[indList[ii]]
                            atMapList.append(dD)

                else:
                    logger.error("+ERROR -unanticipated data container %s", cN)

        except Exception as e:
            logger.error("+ERROR - error processing schema map file %s", schemaMapFile)
            logger.exception("Failing with %s", str(e))

        return tableNameList, atDefList, atMapList, tableAbbrevD, attribAbbrevD


if __name__ == "__main__":
    pass
